c ******************************************************* c
c
      subroutine G_ab_RPT(k, eta1, eta2, G11, G12, G21, G22)
c                                                         c
c                                                         c
c               Time-stamp: <2012-07-31 15:42:18 ataruya> c 
c ******************************************************* c
c
c     Renormalized propagator by Crocce & Scoccimarro (2006b)
c
c     computing correction factor "alpha(z)" has been included
c
      implicit none
c
      integer ikmax, ik_max, ialpha
      parameter(ikmax=5000)
      integer j, jmin, jmax
      real*8 k, eta1, eta2, G11, G12, G21, G22
      real*8 eta, XX, YY, dd2
      real*8 alpha, beta, gamma, delta
      real*8 beta_d, beta_g, gamma_d, gamma_g
      real*8 ak(ikmax), ff(ikmax), gg(ikmax), hh(ikmax), ii(ikmax)
      real*8 fk, gk, hk, ik, s, ds
      real*8 alpha_z
      common /gk_data/ ak, ff, gg, hh, ii, ik_max
      common /alpha_corr/ ialpha
c     -------------------------------------------------
c
c     //////// interpolating one-loop propagator ////////
c
      call hunt(ak, ik_max, k, j)
c
      jmin = j - 2
      jmax = j + 2
      if(jmin.lt.1) jmin = 1
      if(jmax.ge.ik_max) jmax = ik_max
c
      call polint(ak(jmin), ff(jmin), jmax-jmin+1, k, s, ds)
      fk = s
      call polint(ak(jmin), gg(jmin), jmax-jmin+1, k, s, ds)
      gk = s
      call polint(ak(jmin), hh(jmin), jmax-jmin+1, k, s, ds)
      hk = s
      call polint(ak(jmin), ii(jmin), jmax-jmin+1, k, s, ds)
      ik = s
c
c     //////// interpolating alpha ////////
c
      if(ialpha.eq.0) then 
         call find_alpha(eta1, alpha_z)
      else
         alpha_z = 1.d0
      endif
c
      alpha_z = alpha_z**2
c
c     ///////// computing growth functions ////////
c
      eta = eta1 - eta2
c
      call growth_functions(eta, alpha, beta, gamma, delta)
c
      beta_d = beta * dexp(1.5d0*eta)
      beta_g = beta / dexp(eta)
      gamma_d = gamma * dexp(1.5d0*eta)
      gamma_g = gamma / dexp(eta)
c
c     ///////// renormalized propagators ////////
c
      dd2 = dexp(2.d0*eta2)
c
      XX = dd2 * (alpha * fk - beta_g * ik)
      YY = dd2 * (delta * gk - gamma_d * hk)
      G11 = 0.6d0 * dexp(eta) * dexp(alpha_z*XX) 
     &     + 0.4d0 * dexp(-1.5d0*eta) * dexp(alpha_z*YY)
c
      XX = dd2 * (alpha * fk - beta_g * hk)
      YY = dd2 * (delta * fk - gamma_d * hk)
      G12 = 0.4d0 * dexp(eta) * dexp(alpha_z*XX) 
     &     - 0.4d0 * dexp(-1.5d0*eta) * dexp(alpha_z*YY)
c
ccc      XX = dd2 * (alpha * gk + gamma_g * hk)
ccc      YY = dd2 * (delta * gk + gamma_d * ik)
      XX = dd2 * (alpha * gk + gamma_g * ik)
      YY = dd2 * (delta * gk + beta_d * ik)
      G21 = 0.6d0 * dexp(eta) * dexp(alpha_z*XX) 
     &     - 0.6d0 * dexp(-1.5d0*eta) * dexp(alpha_z*YY)
c
      XX = dd2 * (alpha * gk - 1.5d0* gamma_g * ik)
      YY = dd2 * (delta * fk - (2.d0/3.d0)* beta_d * hk)
      G22 = 0.4d0 * dexp(eta) * dexp(alpha_z*XX) 
     &     + 0.6d0 * dexp(-1.5d0*eta) * dexp(alpha_z*YY)
c
      end
c
c ******************************************************* c
c
      subroutine G_ab_Closure(k, eta1, eta2, G11, G12, G21, G22)
c
c ******************************************************* c
c
c     Renormalized propagator by closure theory
c
      implicit none
c
      integer ikmax, ik_max, ialpha
      parameter(ikmax=5000)
      integer j, jmin, jmax
      real*8 k, eta1, eta2, G11, G12, G21, G22
      real*8 eta, XX, YY, dd, bessj1
      real*8 alpha, beta, gamma, delta
      real*8 beta_d, beta_g, gamma_d, gamma_g
      real*8 ak(ikmax), ff(ikmax), gg(ikmax), hh(ikmax), ii(ikmax)
      real*8 fk, gk, hk, ik, s, ds
      real*8 alpha_z
      common /gk_data/ ak, ff, gg, hh, ii, ik_max
      common /alpha_corr/ ialpha
c     -------------------------------------------------
c
c     //////// interpolating one-loop propagator ////////
c
      call hunt(ak, ik_max, k, j)
c
      jmin = j - 2
      jmax = j + 2
      if(jmin.lt.1) jmin = 1
      if(jmax.ge.ik_max) jmax = ik_max
c
      call polint(ak(jmin), ff(jmin), jmax-jmin+1, k, s, ds)
      fk = s
      call polint(ak(jmin), gg(jmin), jmax-jmin+1, k, s, ds)
      gk = s
      call polint(ak(jmin), hh(jmin), jmax-jmin+1, k, s, ds)
      hk = s
      call polint(ak(jmin), ii(jmin), jmax-jmin+1, k, s, ds)
      ik = s
c
c     //////// interpolating alpha ////////
c
      if(ialpha.eq.0) then 
         call find_alpha(eta1, alpha_z)
      else
         alpha_z = 1.d0
      endif
c
c     ///////// computing growth functions ////////
c
      eta = eta1 - eta2
c
      call growth_functions(eta, alpha, beta, gamma, delta)
c
      beta_d = beta * dexp(1.5d0*eta)
      beta_g = beta / dexp(eta)
      gamma_d = gamma * dexp(1.5d0*eta)
      gamma_g = gamma / dexp(eta)
c
c     ///////// renormalized propagators ////////
c
      dd = dexp(eta2) 
c
      XX = dd * alpha_z * dsqrt(2.d0 * dabs(alpha * fk - beta_g * ik))
      YY = dd * alpha_z * dsqrt(2.d0 * dabs(delta * gk - gamma_d * hk))
      
      G11 = 0.6d0 * dexp(eta) * bessj1(2.d0*XX)/XX 
     &     + 0.4d0 * dexp(-1.5d0*eta) * bessj1(2.d0*YY)/YY 
c
      XX = dd * alpha_z * dsqrt(2.d0 * dabs(alpha * fk - beta_g * hk))
      YY = dd * alpha_z * dsqrt(2.d0 * dabs(delta * fk - gamma_d * hk))
      G12 = 0.4d0 * dexp(eta) * bessj1(2.d0*XX)/XX
     &     - 0.4d0 * dexp(-1.5d0*eta) * bessj1(2.d0*YY)/YY
c
ccc      XX = dd * alpha_z * dsqrt(2.d0 * dabs(alpha * gk + gamma_g * hk))
ccc      YY = dd * alpha_z * dsqrt(2.d0 * dabs(delta * gk + gamma_d * ik))
      XX = dd * alpha_z * dsqrt(2.d0 * dabs(alpha * gk + gamma_g * ik))
      YY = dd * alpha_z * dsqrt(2.d0 * dabs(delta * gk + beta_d * ik))

      G21 = 0.6d0 * dexp(eta) * bessj1(2.d0*XX)/XX
     &     - 0.6d0 * dexp(-1.5d0*eta) * bessj1(2.d0*YY)/YY
c
      XX = dd * alpha_z 
     &     * dsqrt(2.d0 * dabs(alpha * gk - 1.5d0* gamma_g * ik))
      YY = dd * alpha_z * 
     &     dsqrt(2.d0 * dabs(delta * fk - 2.d0/3.d0 * beta_d * hk))
      G22 = 0.4d0 * dexp(eta) * bessj1(2.d0*XX)/XX
     &     + 0.6d0 * dexp(-1.5d0*eta) * bessj1(2.d0*YY)/YY
c
      end
c
c ******************************************************* c
c
      subroutine G_ab_linear(eta1, eta2, G11, G12, G21, G22)
c
c ******************************************************* c
c
c     Linear propagator 
c
      implicit none
c
      real*8 eta1, eta2, G11, G12, G21, G22
      real*8 eta
c     -------------------------------------------------
c
c     
      eta = eta1 - eta2
c
      G11 = 0.6d0 * dexp(eta) + 0.4d0 * dexp(-1.5d0*eta) 
c
      G12 = 0.4d0 * dexp(eta) - 0.4d0 * dexp(-1.5d0*eta) 
c
      G21 = 0.6d0 * dexp(eta) - 0.6d0 * dexp(-1.5d0*eta) 
c
      G22 = 0.4d0 * dexp(eta) + 0.6d0 * dexp(-1.5d0*eta)
c
      end
c
c ******************************************************* c
c
      subroutine growth_functions(eta, alpha, beta, gamma, delta)
c
c ******************************************************* c
c
      implicit none
c
      real*8 eta, alpha, beta, gamma, delta
c     ------------------------------------------
c
      alpha = dexp(2.d0*eta) - 1.4d0 * dexp(eta) 
     &        + 0.4d0 * dexp(-1.5d0*eta)
      beta  = 0.6d0 * dexp(2.d0*eta) - dexp(eta) 
     &        + 0.4d0 * dexp(-0.5d0*eta)
      gamma = 0.4d0 * dexp(2.d0*eta) - dexp(0.5d0*eta) 
     &        + 0.6d0 * dexp(-0.5d0*eta)
      delta = 0.4d0 * dexp(3.5d0*eta) - 1.4d0 * dexp(eta) + 1.d0
c
      end
c
c ******************************************************* c
c
      subroutine one_loop_propagator
c
c ******************************************************* c
c
c     computing one-loop propagator 
c     
      implicit none
c
      integer  ik, ikmax, ik_max1, ik_max2
      integer  ix, ix_max
      parameter(ikmax=5000)
      parameter(ix_max=6000)
      real*8   ak(ikmax), pk(ikmax)
      real*8   ak2(ikmax), ff(ikmax), gg(ikmax), hh(ikmax), ii(ikmax)
      real*8   kernel_ff, kernel_gg, kernel_hh, kernel_ii
      real*8   pi, kmin, kmax, xmin, xmax
      real*8   pklink, pklin1, pklin2
      real*8   x1, x2, k
      character infile*50
      common /pk_data/ ak, pk, ik_max1
      common /gk_data/ ak2, ff, gg, hh, ii, ik_max2
      pi = 4.d0 * datan(1.d0)
c     -------------------------------------------------
c
      ik_max2 = ik_max1
c
      kmin = ak(1)
      kmax = ak(ik_max1)
c
      do 10 ik=1, ik_max1
c
      k = ak(ik)
c
      xmin= kmin / k
      xmax= kmax / k
c
      ff(ik) = 0.0
      gg(ik) = 0.0
      hh(ik) = 0.0
      ii(ik) = 0.0
c
      x1 = xmin
      call find_pk(k*x1, pklin1)
c
      do 20 ix=1, ix_max-1
c
         x2 = xmin * (xmax/xmin) ** (dble(ix)/dble(ix_max-1))
         call find_pk(k*x2, pklin2)
c
         ff(ik) = ff(ik) + 
     &        (kernel_ff(x1) * pklin1 + kernel_ff(x2) * pklin2)
     &        * (x2 - x1)/2.d0 
         gg(ik) = gg(ik) + 
     &        (kernel_gg(x1) * pklin1 + kernel_gg(x2) * pklin2)
     &        * (x2 - x1)/2.d0 
         hh(ik) = hh(ik) + 
     &        (kernel_hh(x1) * pklin1 + kernel_hh(x2) * pklin2)
     &        * (x2 - x1)/2.d0 
         ii(ik) = ii(ik) + 
     &        (kernel_ii(x1) * pklin1 + kernel_ii(x2) * pklin2)
     &        * (x2 - x1)/2.d0 
c     
         x1 = x2
         pklin1 = pklin2
 20   continue
c
      ff(ik) =   k**3 / (2.d0*pi*pi) * ff(ik) 
      gg(ik) =   k**3 / (2.d0*pi*pi) * gg(ik) 
      hh(ik) =   k**3 / (2.d0*pi*pi) * hh(ik) 
      ii(ik) =   k**3 / (2.d0*pi*pi) * ii(ik) 
      ak2(ik) = ak(ik)
c
 10   continue
c
      end
c
c ******************************************************* c
c
      function kernel_ff(x)
c
c ******************************************************* c
c
      implicit none
      real*8 kernel_ff, x
c
      if((x.le.10.d0).and. abs(x-1.d0).ge.1.d-2) then
         kernel_ff = 6.d0/x/x - 79.d0 + 50.d0*x*x - 21.d0*x*x*x*x 
     &        + 0.75d0*(1.d0/x-x)**3 * (2.d0 + 7.d0*x*x) 
     &        * dlog(abs((1.d0-x)/(1.d0+x))**2)
         kernel_ff = kernel_ff / 504.d0
      elseif(abs(x-1.d0).lt.1.d-2) then 
         kernel_ff = - 11.d0/126.d0 + (x-1.d0)/126.d0 
     &        - 29.d0/252.d0 * (x-1.d0)**2
      elseif(x.gt.10.d0) then
         kernel_ff = - 61.d0/630.d0 + 2.d0/105.d0/x/x 
     &        - 10.d0/1323.d0 /x/x/x/x
      endif
      end
c
c ******************************************************* c
c
      function kernel_gg(x)
c
c ******************************************************* c
c
      implicit none
      real*8 kernel_gg, x
c
      if((x.le.10.d0).and. abs(x-1.d0).ge.1.d-2) then
         kernel_gg = 6.d0/x/x - 41.d0 + 2.d0*x*x - 3.d0*x*x*x*x 
     &        + 0.75d0*(1.d0/x-x)**3 * (2.d0 + x*x) 
     &        * dlog(abs((1.d0-x)/(1.d0+x))**2)
         kernel_gg = kernel_gg / 168.d0
      elseif(abs(x-1.d0).lt.1.d-2) then 
         kernel_gg = - 3.d0/14.d0 - 5.d0/42.d0*(x-1.d0) 
     &        +  (x-1.d0)**2/84.d0
      elseif(x.gt.10.d0) then
         kernel_gg = - 3.d0/10.d0 + 26.d0/245.d0/x/x 
     &        - 38.d0/2205.d0 /x/x/x/x
      endif
      end
c
c ******************************************************* c
c
      function kernel_hh(x)
c
c ******************************************************* c
c
      implicit none
      real*8 kernel_hh, x
c
      if((x.le.10.d0).and. abs(x-1.d0).ge.1.d-3) then
         kernel_hh = 6.d0/x/x + 1.d0 + 9.d0*x*x*x*x 
     &        + 0.75d0*(1.d0/x-x)**2 * (2.d0/x + 5.d0*x + 3.d0*x*x*x)
     &        * dlog(abs((1.d0-x)/(1.d0+x))**2)
         kernel_hh = kernel_hh / 24.d0
      elseif(abs(x-1.d0).lt.1.d-3) then 
         kernel_hh = 2.d0/3.d0 + (x-1.d0)
      elseif(x.gt.10.d0) then
         kernel_hh = 19.d0/30.d0 + 32.d0/105.d0/x/x 
     &        - 4.d0/21.d0 /x/x/x/x
      endif
      end
c
c ******************************************************* c
c
      function kernel_ii(x)
c
c ******************************************************* c
c
      implicit none
      real*8 kernel_ii, x
c
      if((x.le.10.d0).and. abs(x-1.d0).ge.1.d-3) then
         kernel_ii = 6.d0/x/x + 29.d0 - 18.d0*x*x + 27.d0*x*x*x*x 
     &        + 0.75d0*(1.d0/x-x)**2 * (2.d0/x + 9.d0*x + 9.d0*x*x*x)
     &        * dlog(abs((1.d0-x)/(1.d0+x))**2)
         kernel_ii = - kernel_ii / 72.d0
      elseif(abs(x-1.d0).lt.1.d-3) then 
         kernel_ii = -11.d0/18.d0 - 5.d0/6.d0*(x-1.d0)
      elseif(x.gt.10.d0) then
         kernel_ii = -67.d0/90.d0 + 2.d0/315.d0/x/x 
     &        + 26.d0/315.d0 /x/x/x/x
      endif
      end
c
c ******************************************************* c
c
      FUNCTION bessj1(x)
c
c ******************************************************* c
c
      implicit none
c
      REAL*8 bessj1,x
      REAL*8 ax,xx,z
      DOUBLE PRECISION p1,p2,p3,p4,p5,q1,q2,q3,q4,q5,r1,r2,r3,r4,r5,r6,
     *s1,s2,s3,s4,s5,s6,y
      SAVE p1,p2,p3,p4,p5,q1,q2,q3,q4,q5,r1,r2,r3,r4,r5,r6,s1,s2,s3,s4,
     *s5,s6
      DATA r1,r2,r3,r4,r5,r6/72362614232.d0,-7895059235.d0,
     *242396853.1d0,-2972611.439d0,15704.48260d0,-30.16036606d0/,s1,s2,
     *s3,s4,s5,s6/144725228442.d0,2300535178.d0,18583304.74d0,
     *99447.43394d0,376.9991397d0,1.d0/
      DATA p1,p2,p3,p4,p5/1.d0,.183105d-2,-.3516396496d-4,
     *.2457520174d-5,-.240337019d-6/, q1,q2,q3,q4,q5/.04687499995d0,
     *-.2002690873d-3,.8449199096d-5,-.88228987d-6,.105787412d-6/
      if(abs(x).lt.8.)then
        y=x**2
        bessj1=x*(r1+y*(r2+y*(r3+y*(r4+y*(r5+y*r6)))))/(s1+y*(s2+y*(s3+
     *y*(s4+y*(s5+y*s6)))))
      else
        ax=abs(x)
        z=8./ax
        y=z**2
        xx=ax-2.356194491
        bessj1=sqrt(.636619772/ax)*(cos(xx)*(p1+y*(p2+y*(p3+y*(p4+y*
     *p5))))-z*sin(xx)*(q1+y*(q2+y*(q3+y*(q4+y*q5)))))*sign(1.d0,x)
      endif
      return
      END
C  (C) Copr. 1986-92 Numerical Recipes Software z!0(0.
c
c ******************************************************* c
c
